from typing import List
import statistics as st
import itertools

MODE_OF_EMPTY_LIST = "Nothing to see here"
TYPE_NUMERIC = "numeric"
TYPE_NOMINAL = "nominal"
TYPE_TYPE = "type"
TYPE_TUPLE = "tuple"
TYPE_SAME_AS_INPUT = "as input"


class Aggregator:
    def __init__(self, name):
        self.name = name
        self.input_types = set()
        self.output_type = None
        self.is_projection = False

    def __repr__(self):
        return self.name

    def get_name(self):
        return self.name

    def aggregate_flat(self, a_list):
        raise NotImplementedError("Should be implemented by a subclass!")

    def aggregate(self, ls):
        return [self.aggregate_flat(a_list) for a_list in ls]

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __lt__(self, other):
        return self.get_name() < other.get_name()


class Count(Aggregator):
    def __init__(self):
        super().__init__("count")
        self.input_types = {TYPE_NUMERIC, TYPE_NOMINAL, TYPE_TYPE, TYPE_TUPLE}
        self.output_type = TYPE_NUMERIC

    def aggregate_flat(self, a_list):
        return len(a_list)


class CountUnique(Aggregator):
    def __init__(self):
        super().__init__("countUnique")
        self.input_types = {TYPE_NUMERIC, TYPE_NOMINAL, TYPE_TYPE,
                            TYPE_TUPLE}  # Should numeric be removed?
        self.output_type = TYPE_NUMERIC

    def aggregate_flat(self, a_list):
        return len(set(a_list))


class Min(Aggregator):
    def __init__(self):
        super().__init__("min")
        self.input_types = {TYPE_NUMERIC}
        self.output_type = TYPE_NUMERIC

    def aggregate_flat(self, a_list):
        return min(a_list) if a_list else float("inf")


class Max(Aggregator):
    def __init__(self):
        super().__init__("max")
        self.input_types = {TYPE_NUMERIC}
        self.output_type = TYPE_NUMERIC

    def aggregate_flat(self, a_list):
        return max(a_list) if a_list else float("-inf")


class Mean(Aggregator):
    def __init__(self):
        super().__init__("mean")
        self.input_types = {TYPE_NUMERIC}
        self.output_type = TYPE_NUMERIC

    def aggregate_flat(self, a_list):
        return st.mean(a_list) if a_list else float(
            "inf")  # slightly better than 0? None?


class Sum(Aggregator):
    def __init__(self):
        super().__init__("sum")
        self.input_types = {TYPE_NUMERIC}
        self.output_type = TYPE_NUMERIC

    def aggregate_flat(self, a_list):
        return sum(a_list)


class Mode(Aggregator):
    def __init__(self):
        super().__init__("mode")
        self.input_types = {TYPE_NOMINAL, TYPE_TYPE}
        self.output_type = TYPE_SAME_AS_INPUT

    def aggregate_flat(self, a_list):
        d = {}
        for j in a_list:
            if j not in d:
                d[j] = 1
            else:
                d[j] += 1
        if d:
            return max(sorted(d),
                       key=lambda t: d[t])  # Every time the same element
        else:
            return MODE_OF_EMPTY_LIST  # float("inf")


class Flatten(Aggregator):
    def __init__(self):
        super().__init__("flatten")
        self.input_types = {TYPE_NUMERIC, TYPE_NOMINAL, TYPE_TYPE}
        self.output_type = TYPE_SAME_AS_INPUT

    def aggregate(self, ls):
        return list(itertools.chain.from_iterable(ls))

    def aggregate_flat(self, a_list):
        return a_list


class FlattenUnique(Aggregator):
    def __init__(self):
        super().__init__("flattenUnique")
        self.input_types = {TYPE_NUMERIC, TYPE_NOMINAL, TYPE_TYPE}
        self.output_type = TYPE_SAME_AS_INPUT

    def aggregate(self, ls):
        return list(set(itertools.chain.from_iterable(ls)))

    def aggregate_flat(self, a_list):
        return a_list


class Project(Aggregator):
    def __init__(self, component, aggregator):
        super().__init__("projection{}".format(component))
        self.input_types = {TYPE_TUPLE}
        self.output_type = "unknown"
        self.is_projection = True
        self.component = component
        self.aggregator = aggregator
        if self.aggregator is not None:
            self.name += "_" + self.aggregator.name
            self.output_type = self.aggregator.output_type

    def aggregate_flat(self, a_list):
        return self.aggregator.aggregate_flat(
            [e[self.component] for e in a_list])


class Project0(Project):
    def __init__(self, aggregator):
        super().__init__(0, aggregator)


class Project1(Project):
    def __init__(self, aggregator):
        super().__init__(1, aggregator)


class Project2(Project):
    def __init__(self, aggregator):
        super().__init__(2, aggregator)


class Project3(Project):
    def __init__(self, aggregator):
        super().__init__(3, aggregator)


class Project4(Project):
    def __init__(self, aggregator):
        super().__init__(4, aggregator)


class Project5(Project):
    def __init__(self, aggregator):
        super().__init__(5, aggregator)


class Project6(Project):
    def __init__(self, aggregator):
        super().__init__(6, aggregator)


class Project7(Project):
    def __init__(self, aggregator):
        super().__init__(7, aggregator)


FLATTEN = Flatten()
FLATTEN_UNIQUE = FlattenUnique()
COUNT = Count()
COUNT_UNIQUE = CountUnique()
MIN = Min()
MAX = Max()
MEAN = Mean()
SUM = Sum()
MODE = Mode()
PROJECT = Project("", None)
PROJECTIONS = [
    Project0, Project1, Project2, Project3, Project4, Project5, Project6,
    Project7
]

ALL_AGGREGATORS = [
    FLATTEN, FLATTEN_UNIQUE, COUNT, COUNT_UNIQUE, MIN, MAX, MEAN, SUM, MODE,
    PROJECT
]  # type: List['Aggregator']
NUMERIC_AGGREGATORS = [
    a for a in ALL_AGGREGATORS if TYPE_NUMERIC in a.input_types
]  # type: List['Aggregator']
NOMINAL_AGGREGATORS = [
    a for a in ALL_AGGREGATORS
    if TYPE_NOMINAL in a.input_types and TYPE_TYPE in a.input_types
]  # type: List['Aggregator']
COMPLEX_ENOUGH_AGGREGATORS = [
    a for a in ALL_AGGREGATORS if TYPE_TUPLE in a.input_types
]
CRITICAL_VALUES = [float("inf"), float("-inf"), MODE_OF_EMPTY_LIST]



class Variable:
    def __init__(self, name: str, value_type, value):
        self.name = name
        self.value_type = value_type
        self.value = value

    def __repr__(self):
        v = self.value if not self.is_unset() else "?"
        return "{}({})".format(self.name, v)

    def __eq__(self, other):
        return self.name == other.name

    def __lt__(self, other):
        return self.name < other.name

    def __hash__(self):
        return hash(self.name)

    def rename(self, n):
        self.name = n

    def get_name(self):
        return self.name

    def set_value(self, v):
        self.value = v

    def unset_value(self):
        self.value = None

    def is_unset(self):
        return self.value is None

    def get_value(self):
        return self.value

    def is_constant(self):
        return False

    def can_vary(self):
        return False

    def make_copy(self):
        raise NotImplementedError("This should be implemented by a subclass.")


class ConstantVariable(Variable):
    def __init__(self, name, variable_type, value):
        super().__init__(name, variable_type, value)
        assert self.name.startswith("C")

    def is_constant(self):
        return True

    def make_copy(self):
        return ConstantVariable(self.name, self.value_type, self.value)


class VariableVariable(Variable):
    def __init__(self, name, variable_type, value):
        super().__init__(name, variable_type, value)
        assert self.name[0] in "XY"  # X for the initial ones, Y for the others

    def can_vary(self):
        return True

    def make_copy(self):
        return VariableVariable(self.name, self.value_type, self.value)



from ...data.data_and_statistics import Dataset, Datum
from .variables import Variable
# from relation import Relation
from typing import Dict, List, Tuple, Union
from py4j.java_collections import ListConverter, MapConverter
from ...data.relation import Relation
from .aggregators import Aggregator


def send_data(data: Dataset, client, wrapper):
    def l_convert(l0):
        return list_converter.convert(l0, client)

    def m_convert(m0):
        return map_converter.convert(m0, client)

    list_converter = ListConverter()
    map_converter = MapConverter()

    # send target_data: we can send only descriptive parts of the data
    simplified_target = []
    for d in data:
        simplified_target.append([d.identifier])
        simplified_target[-1] += list(d.get_descriptive())
        simplified_target[-1] = l_convert(simplified_target[-1])
    simplified_target = l_convert(simplified_target)
    # target data
    wrapper.load_target_data(simplified_target)
    # relations
    relations = {
        r.name: l_convert(r.get_types())
        for r in data.get_descriptive_data().values()
    }
    wrapper.load_relations(m_convert(relations), data.data_file)

    # wrapper.load(l_convert([m_convert({2: "we", 3: "d"}), m_convert({"3": 2, "21": 21})]))


def send_variables(example: Dict[str, Variable], client, wrapper):
    def l_convert(l0):
        return list_converter.convert(l0, client)

    list_converter = ListConverter()

    variables = []
    for variable in example.values():
        v_name = variable.get_name()
        v_type = variable.value_type
        v_value = variable.get_value()
        v_can_vary = variable.can_vary()
        variables.append(l_convert([v_name, v_type, v_value, v_can_vary]))
    wrapper.load_variables(l_convert(variables))


def compute_test_values(target_data: List[Datum],
                        target_relation_vars: List[str],
                        rc_modified: List[Tuple[Relation, List[str]]],
                        filtered_agg_chains: List[List[Aggregator]],
                        r_key: Tuple[Tuple[str, Tuple[bool],
                                           Tuple[Union[str, int, Variable]]]],
                        a_keys: List[Tuple[str]], nb_fresh_vars: int,
                        fresh_indices: List[int],
                        known_unknown: List[List[List[int]]], client, wrapper):
    def l_convert(l0):
        return list_converter.convert(l0, client)

    # def m_convert(m0):
    #     return map_converter.convert(m0, client)

    list_converter = ListConverter()
    # map_converter = MapConverter()

    # set_example_values_time = 0
    # find_values_time = 0

    # conversion
    target_data_converted = l_convert([d.identifier for d in target_data])

    target_relation_vars_converted = l_convert(target_relation_vars)

    rc_modified_converted = []
    for r, variables in rc_modified:
        rc_modified_converted.append(l_convert([r.get_name()] + variables))
    rc_modified_converted = l_convert(rc_modified_converted)

    filtered_agg_chains_converted = []
    for aggregators in filtered_agg_chains:
        filtered_agg_chains_converted.append(
            l_convert([a.get_name() for a in aggregators]))
    filtered_agg_chains_converted = l_convert(filtered_agg_chains_converted)

    if r_key is None:
        r_key_converted = None
    else:
        r_key_converted = []
        for name, booleans, variables in r_key:
            n = len(booleans)
            assert n == len(variables)
            new_element = ["" for _ in range(1 + 2 * n)]
            new_element[0] = name
            for i in range(n):
                new_element[1 + i] = str(booleans[i]).lower()
                new_element[1 + n + i] = variables[i]
            r_key_converted.append(l_convert(new_element))
        r_key_converted = l_convert(r_key_converted)

    a_keys_converted = []
    for t in a_keys:
        a_keys_converted.append(l_convert(list(t)))
    a_keys_converted = l_convert(a_keys_converted)

    fresh_indices_converted = l_convert(fresh_indices)

    known_unknown_converted = []
    for known, unknown in known_unknown:
        known_unknown_converted.append(
            l_convert([l_convert(known), l_convert(unknown)]))
    known_unknown_converted = l_convert(known_unknown_converted)

    values = wrapper.compute_test_values(
        target_data_converted, target_relation_vars_converted,
        rc_modified_converted, filtered_agg_chains_converted, r_key_converted,
        a_keys_converted, nb_fresh_vars, fresh_indices_converted,
        known_unknown_converted)
    all_test_values = [[x for x in part] for part in values]
    return all_test_values



from typing import Set, Tuple, List, Union
import re
from ..utilities.my_utils import *
from ..learners.core.variables import Variable
from ..utilities.my_exceptions import WrongValueException
import numpy as np


class Relation:
    constant_nominal = "nominal"
    constant_numeric = "numeric"
    constant_multi_target = "multi_target["
    relation_type_constant = [
        constant_nominal, constant_numeric, constant_multi_target
    ]
    allowed_chars = "-.,+ A-Za-z0-9_\\[\\]"
    # constant_type = "constant"
    tuple_pattern = "{{}}\\(([{} ,]+)\\)".format(allowed_chars)
    time_efficient_search_bound = 3

    def __init__(self, name: str, related_objects: Union[Set[Tuple[str]],
                                                         None],
                 file: Union[str, None], types):
        self.name = name
        self.all_tuples = set()
        self.types = types
        self.different_values = [-1] * len(self.types)
        self.arity = len(self.types)
        self.all_tuples_by_subsets = {}
        self.init_all_tuples_by_subsets()
        self.file = file
        p1 = related_objects is None
        p2 = self.file is None
        assert p1 + p2 == 1
        if p1:
            with open(file) as f:
                for whole_line in f:
                    line = whole_line[:whole_line.find("//")]  # comments
                    if line:
                        self.try_add_tuple(line)
        else:
            self.all_tuples = related_objects
            # TODO: This does not update tuples by subsets?

    def __repr__(self):
        set_part = []
        total_len = 0
        for t in self.all_tuples:
            if total_len > 30:
                set_part.append("...")
                break
            set_part.append(str(t))
            total_len += len(set_part[-1])
        return "Relation({}, {{{}}})".format(self.name, ", ".join(set_part))

    def __eq__(self, other):
        return self.name == other.name  # and self.types == other.types and self.all_tuples == other.all_tuples

    def should_use_tuples_by_subsets(self):
        return 1 <= self.arity <= Relation.time_efficient_search_bound

    def init_all_tuples_by_subsets(self):
        if self.should_use_tuples_by_subsets():
            pattern = "{{:0>{}b}}".format(self.arity)
            subset_codes = [
                pattern.format(i) for i in range(1, 2**self.arity - 1)
            ]
            for subset_code in subset_codes:
                self.all_tuples_by_subsets[subset_code] = {}
            for t in self.all_tuples:
                self.try_add_one_to_tuples_by_subsets(t)

    def try_add_tuple(self, line: str):
        """
        Converts 'r(x,y)' to (x, y) and adds it to all tuples.

        :param line: a string of form <relation name>(obj1, obj2, ...)
        """
        related_list = parse_relation_arguments(line, self.name)
        relation_tuple = tuple(
            Relation.intelligent_parse(v_type, v_value)
            for v_type, v_value in zip(self.types, related_list))
        self.add_parsed_tuple(relation_tuple)

    def add_parsed_tuple(self, t):
        self.all_tuples.add(t)
        if self.should_use_tuples_by_subsets():
            self.try_add_one_to_tuples_by_subsets(t)

    def try_add_one_to_tuples_by_subsets_old(self, relation_tuple):
        pattern = "{{:0>{}b}}".format(self.arity)
        subset_codes = [pattern.format(i) for i in range(1, 2**self.arity - 1)]
        for subset_code in subset_codes:
            subset_dict = self.all_tuples_by_subsets[subset_code]
            key_part = []
            value_part = []
            for tuple_component, code_component in zip(relation_tuple,
                                                       subset_code):
                if code_component == "1":
                    key_part.append(tuple_component)
                else:
                    value_part.append(tuple_component)
            key_part = tuple(key_part)
            value_part = tuple(value_part)
            if key_part not in subset_dict:
                subset_dict[key_part] = set()
            subset_dict[key_part].add(value_part)

    def try_add_one_to_tuples_by_subsets(self, relation_tuple):
        pattern = "{{:0>{}b}}".format(self.arity)
        subset_codes = [pattern.format(i) for i in range(1, 2**self.arity - 1)]
        for subset_code in subset_codes:
            subset_dict = self.all_tuples_by_subsets[subset_code]
            key_part = []
            # value_part = []
            for tuple_component, code_component in zip(relation_tuple,
                                                       subset_code):
                if code_component == "1":
                    key_part.append(tuple_component)
                # else:
                #     value_part.append(tuple_component)
            key_part = tuple(key_part)
            value_part = relation_tuple  # tuple(value_part)
            if key_part not in subset_dict:
                subset_dict[key_part] = []  # set()
            # subset_dict[key_part].add(value_part)
            subset_dict[key_part].append(value_part)

    def get_all_old(self, variables: List[Variable],
                    known_values: List[int]) -> List[Tuple[Variable]]:
        """
        :param variables: e.g., [X0(2.1), X1(?), X2('b')]
        :param known_values: list of indices of the variables that have known value, e.g., [0, 2]
        :return: list of tuples that satisfy the constraints, e.g., all triplets (x, y, z), for which
          x == 2.1 and z = 'b'.
        """
        def check_ok(related):
            for j in known_values:
                if variables[j].get_value() != related[j]:
                    return False
            return True

        def merge(known_part, unknown_part):
            merged = [None] * len(subset_code)
            i_known = 0
            i_unknown = 0
            for j in range(len(merged)):
                if subset_code[j] == "1":
                    merged[j] = known_part[i_known]
                    i_known += 1
                else:
                    merged[j] = unknown_part[i_unknown]
                    i_unknown += 1
            return tuple(merged)

        if not self.should_use_tuples_by_subsets():
            return [t for t in self.all_tuples if check_ok(t)]
        else:
            key_part = tuple([variables[i].get_value() for i in known_values])
            if len(known_values) == self.arity:
                return [key_part] if key_part in self.all_tuples else []
            elif len(known_values) == 0:
                return list(self.all_tuples)
            else:
                subset_code = ["0"] * self.arity
                for i in known_values:
                    subset_code[i] = "1"
                subset_code = "".join(subset_code)
                subset_dict = self.all_tuples_by_subsets[subset_code]
                if key_part in subset_dict:
                    value_parts = subset_dict[key_part]
                else:
                    value_parts = set()
                return [
                    merge(key_part, value_part) for value_part in value_parts
                ]

    def get_all(self, variables: List[Variable],
                known_values: List[int]) -> List[Tuple[Variable]]:
        """
        :param variables: e.g., [X0(2.1), X1(?), X2('b')]
        :param known_values: list of indices of the variables that have known value, e.g., [0, 2]
        :return: list of tuples that satisfy the constraints, e.g., all triplets (x, y, z), for which
          x == 2.1 and z = 'b'.
        """
        def check_ok(related):
            for j in known_values:
                if variables[j].get_value() != related[j]:
                    return False
            return True

        if not self.should_use_tuples_by_subsets():
            return [t for t in self.all_tuples if check_ok(t)]
        else:
            key_part = tuple([variables[i].get_value() for i in known_values])
            if len(known_values) == self.arity:
                return [key_part] if key_part in self.all_tuples else []
            elif len(known_values) == 0:
                return list(self.all_tuples)
            else:
                subset_code = ["0"] * self.arity
                for i in known_values:
                    subset_code[i] = "1"
                subset_code = "".join(subset_code)
                subset_dict = self.all_tuples_by_subsets[subset_code]
                if key_part in subset_dict:
                    return subset_dict[key_part]
                else:
                    return []

    def get_all_values(self, position):
        return sorted({t[position] for t in self.all_tuples})

    def get_nb_all_values(self, position):
        if self.different_values[position] < 0:
            self.different_values[position] = len(
                self.get_all_values(position))
        return self.different_values[position]

    def get_name(self):
        return self.name

    def get_types(self):
        return self.types

    @staticmethod
    def intelligent_parse(variable_type, variable_value):
        if Relation.is_numeric_type(variable_type):
            # constant numeric
            return try_convert_to_number(variable_value)
        elif Relation.is_nominal_type(variable_type):
            # constant nominal
            return variable_value
        elif Relation.is_multi_target_type(variable_type):
            # multi_target[some type]
            type_of_targets = Relation.get_inner_type_of_multi_target(
                variable_type)
            if not (variable_value.startswith('[')
                    and variable_value.endswith(']')):
                raise WrongValueException(
                    "Multi-target value {} wrongly specified!".format(
                        variable_value))
            target_values_string = intelligent_split(variable_value[1:-1])
            target_values = [
                Relation.intelligent_parse(type_of_targets, v)
                for v in target_values_string
            ]
            if Relation.is_numeric_type(type_of_targets):
                return np.array(target_values)
        else:
            # user-defined
            return variable_value

    @staticmethod
    def is_constant_type(variable_type):
        return Relation.is_nominal_type(variable_type) or Relation.is_numeric_type(variable_type) or\
               Relation.is_multi_target_type(variable_type)

    @staticmethod
    def is_nominal_type(variable_type):
        return variable_type.startswith(Relation.constant_nominal)

    @staticmethod
    def is_numeric_type(variable_type):
        return variable_type.startswith(Relation.constant_numeric)

    @staticmethod
    def is_multi_target_type(variable_type):
        return variable_type.startswith(Relation.constant_multi_target)

    @staticmethod
    def get_inner_type_of_multi_target(variable_type):
        assert Relation.is_multi_target_type(variable_type)
        i0, i1 = variable_type.find('['), variable_type.rfind(']')
        if min(i0, i1) < 0:
            raise WrongValueException(
                "Multi-target type {} wrongly specified!".format(
                    variable_type))
        return variable_type[i0 + 1:i1].strip()

    @staticmethod
    def check_has_ok_types(relation_types):
        for t in relation_types:
            if t[0] == t[0].lower() and not Relation.is_constant_type(t):
                message = "Type {} should either start with a capital letter, " \
                          "or be of form <element><appendix>, where <element> is in {}."
                raise WrongValueException(
                    message.format(t, Relation.relation_type_constant))


def parse_relation_arguments(line: str, relation_name: str):
    assert line.startswith(relation_name)
    tuple_pattern = Relation.tuple_pattern.format(relation_name)
    try:
        related_str = re.search(tuple_pattern, line).group(1).strip()
    except AttributeError:
        print("Relation {}: tuple {} does not match pattern {}".format(
            relation_name, line, tuple_pattern))
        raise
    related_list = intelligent_split(related_str)
    object_name = "^[{}]+$".format(Relation.allowed_chars)
    for o in related_list:
        if re.match(object_name, o) is None:
            if o == "":
                print("Empty string detected in {}".format(line))
            else:
                raise WrongValueException(
                    "{} in {} contains forbidden characters or is empty.".
                    format(o, related_str))
    return related_list


def intelligent_split(line):
    inside_list = False
    inside_quotation = False
    elements = []
    start = 0
    for i, c in enumerate(line):
        if c == '"':
            if i > 0 and line[i - 1] == '\\':
                pass  # escaped quotation
            else:
                inside_quotation = not inside_quotation
        if inside_quotation:
            continue
        if c == '[':
            inside_list = True
        elif c == ']':
            inside_list = False
        elif c == ',' and not inside_list:
            elements.append(line[start:i])
            start = i + 1
    elements.append(line[start:])
    return [e.strip() for e in elements]


def parse_relation_name(line):
    name_pattern = "([{}]+)\\(".format(Relation.allowed_chars)
    try:
        return re.match(name_pattern, line).group(1)
    except AttributeError:
        print(f"Pattern {name_pattern} did not match {line}")
        raise


def parse_relation(line):
    relation_name = parse_relation_name(line)
    related_list = parse_relation_arguments(line, relation_name)
    return relation_name, related_list



import random


class EnsembleRandomGenerator:
    upper_bound = 12345

    def __init__(self, random_seed):
        self.meta_random = random.Random(random_seed)
        self.tree_seeds = random.Random(self.next_seed())
        self.bootstrap_seeds = random.Random(self.next_seed())
        self.sample_rows_seeds = random.Random(self.next_seed())

    def next_seed(self):
        return int(self.meta_random.random() *
                   EnsembleRandomGenerator.upper_bound)

    def next_tree_seed(self):
        return int(self.tree_seeds.random() *
                   EnsembleRandomGenerator.upper_bound)

    def next_bootstrap_seed(self):
        return int(self.bootstrap_seeds.random() *
                   EnsembleRandomGenerator.upper_bound)

    def next_sample_rows_seed(self):
        return int(self.sample_rows_seeds.random() *
                   EnsembleRandomGenerator.upper_bound)


def try_convert_to_number(s):
    # try:
    #     return int(s)
    # except ValueError:
    #     pass
    try:
        return float(s)
    except ValueError:
        pass
    return s


def union_of_two_dicts(d1, d2):
    """
    Computes the union of two dictionaries whose values are lists.
    :param d1: dictionary, e.g.,  {'Person': {'X1', 'X2'}, 'Animal': {'Y1'}}
    :param d2: same as d1, e.g., {'Person': {'X4'}, 'House': {'X3'}}
    :return: a union of dictionaries, e.g., {'Person': {'X1', 'X2', 'X4'}, 'Animal': {'Y1'}, 'House': {'X3'}}
    """
    d = {k: {n for n in v} for k, v in d1.items()}
    for k, v in d2.items():
        if k not in d:
            d[k] = v
        else:
            d[k] |= v
    return d


def average_of_dictionaries(ds):
    n = len(ds)
    together = {}
    # sum
    for d in ds:
        for k, v in d.items():
            if k not in together:
                together[k] = 0.0
            together[k] += v
    # division
    if n > 1:
        for k in together:
            together[k] /= n
    return together


def canonic_sequences_of_new_variables(n):
    """
    For example, if n == 5, we would get
    1 1 1 1 1
    1 1 1 1 2
    1 1 1 2 1
    ...
    1 2 1 2 1
    ...
    but not 2 2 2 2 2, 2 1 2 1 2 etc.
    :param n: The length of the sequence
    :return: All possible sequences of indices of new variables.
    """
    def helper(k):
        if k == 1:
            yield [1], 1
        else:
            for s, m in helper(k - 1):
                for i in range(1, m + 2):
                    yield s + [i], max(i, m)

    if n == 0:
        yield []
    else:
        for t, _ in helper(n):
            yield t


def subsets_of_list(a_list, max_nb_generated=None):
    n = len(a_list)
    nb_subsets = 2**n
    pattern = "{{:0>{}b}}".format(n)
    if max_nb_generated is None:
        max_nb_generated = nb_subsets
    bound = min(nb_subsets, max_nb_generated)
    for i in range(bound):
        s = pattern.format(i)
        chosen_and_not = [[], []]
        for c, e in zip(s, a_list):
            chosen_and_not[c == "0"].append(e)
        yield chosen_and_not


def generalized_counting(a_list, k):
    n = len(a_list)
    indices = [0] * k
    if n > 0:
        while True:
            yield [a_list[i] for i in indices]
            which = k - 1
            while which >= 0 and indices[which] == n - 1:
                which -= 1
            if which < 0:
                break
            for after in range(which + 1, k):
                indices[after] = 0
            indices[which] += 1
    elif k == 0:
        yield []


def get_an_element_of_set(s):
    return next(iter(s))


def arg_max(values):
    m = -float("inf")
    chosen = None
    for i, v in enumerate(values):
        if v > m:
            m = v
            chosen = i
    return chosen


def float_as_string(x, significant_places):
    float_to_str_pattern = "{{:.{}e}}".format(significant_places)
    x_str = float_to_str_pattern.format(float(x))
    # Different lengths can occur:
    # 1) '-' sign at the beginning
    # 2) Different lengths of exponent: we assume that it is either e+12 or e+100
    has_minus = x_str.startswith('-')
    exponent_len = len(x_str) - x_str.find("e") - 2  # either 2 or 3
    start = "" if has_minus else " "
    end = "" if exponent_len == 3 else " "
    return start + x_str + end
